"""
SageMaker metrics by CloudWatch.

"""
from enum import Enum, unique


@unique
class MetricMode(Enum):
    TRAINING = "Training"
    DEVELOP = "Develop"


@unique
class MetricUnit(Enum):
    """
    Inexhaustive list of common CloudWatch metric units

    """
    PERCENT = "Percent"
    COUNT = "Count"
    NONE = "None"


class Metric:
    """
    Record of a metric value at one point in time.

    """
    def __init__(self, name, value, unit=MetricUnit.NONE):
        self.name = name
        self.value = value
        self.unit = unit

        if not isinstance(unit, MetricUnit):
            raise ValueError("Metric needs to be of a supported enum")



from json import dumps

from boto3 import client
from botocore.exceptions import ClientError, NoCredentialsError, NoRegionError

from microcosm_sagemaker.metrics.models import MetricMode


class SageMakerMetrics:
    def __init__(self, graph):
        pass

    def create(self, model_name, metrics=[], hyperparameters={}, mode=MetricMode.TRAINING):
        # Metric dimensions allow us to analyze metric performance against the
        # hyperparameters of our model
        dimensions = [
            {
                "Name": key,
                "Value": value,
            }
            for key, value in hyperparameters.items()
        ]

        metric_data = [
            {
                "MetricName": f"{mode.value} {metric.name}",
                "Dimensions": dimensions,
                "Value": metric.value,
                "Unit": metric.unit.value,
                "StorageResolution": 1
            }
            for metric in metrics
        ]

        try:
            cloudwatch = client("cloudwatch")
            response = cloudwatch.put_metric_data(
                Namespace="/aws/sagemaker/" + model_name,
                MetricData=metric_data,
            )
        except (ClientError, NoCredentialsError, NoRegionError):
            print("CloudWatch publishing disabled")  # noqa: T003
            print(dumps(metric_data, indent=4))  # noqa: T003
            response = None

        return response



import shutil
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Dict

from allennlp.commands.train import train_model
from allennlp.common import Params
from allennlp.models.archival import load_archive
from allennlp.predictors.predictor import Predictor

import microcosm_sagemaker.frameworks.allennlp.vanilla_predictor  # noqa
from microcosm_sagemaker.artifact import BundleInputArtifact, BundleOutputArtifact
from microcosm_sagemaker.bundle import Bundle
from microcosm_sagemaker.frameworks.allennlp.constants import ARTIFACT_NAME, CUDA_DEVICE
from microcosm_sagemaker.input_data import InputData


class AllenNLPBundle(Bundle):
    """
    Higher-order AllenNLP component that can wrap other models, serializing
    our configuration format the way that they expect.

    Note that any paths which appear in allennlp_parameters are expected to be
    relative to the `input_data` directory.

    """
    # To specify custom predictor
    predictor_name: str = "vanilla_predictor"
    allennlp_parameters: Dict[str, Any]

    def fit(self, input_data: InputData) -> None:
        allennlp_params = Params(self.allennlp_parameters)
        self.temporary_allennlp_dir = TemporaryDirectory()
        self.temporary_allennlp_path = Path(self.temporary_allennlp_dir.name)
        with input_data.cd():
            train_model(
                allennlp_params,
                self.temporary_allennlp_path,
            )

        self._set_predictor(self.temporary_allennlp_path)

    def save(self, output_artifact: BundleOutputArtifact) -> None:
        allen_nlp_path = self._allenlp_path(output_artifact.path)
        allen_nlp_path.mkdir(parents=True)

        for child in self.temporary_allennlp_path.iterdir():
            # NB: It is important to use `shutil.move` here rather than
            # `rename` because the former works across filesystems, and
            # SageMaker uses a different filesystem for its temporary
            # directories and the output artifact directory
            shutil.move(child, allen_nlp_path / child.name)

        self.temporary_allennlp_dir.cleanup()

    def load(self, input_artifact: BundleInputArtifact) -> None:
        self._set_predictor(self._allenlp_path(input_artifact.path))

    def _set_predictor(self, allennlp_path: Path) -> None:
        weights_path = allennlp_path / ARTIFACT_NAME

        archive = load_archive(
            allennlp_path / "model.tar.gz",
            weights_file=weights_path,
            cuda_device=CUDA_DEVICE,
        )

        self.predictor = Predictor.from_archive(archive, self.predictor_name)

    def _allenlp_path(self, artifact_path: Path):
        return Path(artifact_path) / "allennlp"



from abc import ABCMeta
from collections import namedtuple

import numpy as np
import pandas as pd

from tabulate import tabulate


MetricPoint = namedtuple("MetricPoint", "key value mode epoch")


class MetricMixin(metaclass=ABCMeta):
    """
    Mixin to add standardized support for logging performance during training/evaluation.

    """
    def __init__(self):
        self.metrics = []

    def add_metric(self, key, value, mode, epoch):
        self.metrics.append(MetricPoint(key, value, mode, epoch))

    def log_epoch_stats(self, epoch):
        """
        Log a formatted version of the stats for this epoch, across all modes
        """
        epoch_metrics = [metric for metric in self.metrics if metric.epoch == epoch]
        self._format_metrics(epoch_metrics)

    def get_best_metric(self, key, mode, best_fn):
        filtered_metrics = [
            metric
            for metric in self.metrics
            if metric.key == key and metric.mode == mode
        ]
        return best_fn(filtered_metrics, key=lambda metric: metric.value)

    def log_best_stats(self, key, mode, best_fn=max):
        """
        Log the best epoch stats given the desired key & mode
        """
        # Get the best metric
        best_metric = self.get_best_metric(key, mode, best_fn)

        print(f"Best Epoch ({key}, {mode}) = {best_metric.epoch}")  # noqa: N123

        epoch_metrics = [
            metric
            for metric in self.metrics
            if metric.epoch == best_metric.epoch
        ]
        self._format_metrics(epoch_metrics)

    def _format_metrics(self, metrics):
        rows = {metric.key for metric in metrics}
        cols = {metric.mode for metric in metrics}

        # Sort the sets to guarantee alphanumeric output
        rows = sorted(list(rows))
        cols = sorted(list(cols))

        # Format each row based on the key/mode configuration
        grid = np.empty((len(rows), len(cols)), dtype=float)

        for metric in metrics:
            grid[rows.index(metric.key)][cols.index(metric.mode)] = metric.value

        df = pd.DataFrame(grid, index=rows, columns=cols)

        print(tabulate(df, headers="keys"))  # noqa: N123
